@@ -4,6 +4,7 @@ import logging
 import socket
 from array import array
 import pprint
+from threading import Event, Thread
 
 from .method_framing import MethodReader, MethodWriter
 from .serialization import AMQPWriter
@@ -26,32 +27,19 @@ LIBRARY_PROPERTIES = {
     'capabilities': {},
 }
 
-log = logging.getLogger('amqpy')
+log = logging.getLogger('amqpy.connection')
 
 
 class Connection(AbstractChannel):
     """The connection class provides methods for a client to establish a network connection to a server, and for both
     peers to operate the connection thereafter
     """
-    #: Final heartbeat interval value (in float seconds) after negotiation
-    heartbeat = None
 
-    #: Original heartbeat interval value proposed by client
-    client_heartbeat = None
-
-    #: Original heartbeat interval proposed by server
-    server_heartbeat = None
-
-    #: Time of last heartbeat sent (in monotonic time, if available)
-    last_heartbeat_sent = 0
-
-    #: Time of last heartbeat received (in monotonic time, if available)
-    last_heartbeat_received = 0
-
-    def __init__(self, host='localhost', port=5672, userid='guest', password='guest', login_method='AMQPLAIN',
-                 virtual_host='/', locale='en_US', client_properties=None, ssl=None, connect_timeout=None,
-                 channel_max=65535, frame_max=131072, heartbeat=0, on_blocked=None, on_unblocked=None,
-                 enable_publisher_ack=False):
+    def __init__(self, *, host='localhost', port=5672, ssl=None, connect_timeout=None,
+                 userid='guest', password='guest', login_method='AMQPLAIN', virtual_host='/', locale='en_US',
+                 channel_max=65535, frame_max=131072, heartbeat=0, auto_heartbeat=False,
+                 enable_publisher_ack=False, client_properties=None,
+                 on_blocked=None, on_unblocked=None):
         """Create a connection to the specified host
 
         If you are using SSL, make sure the correct port number is specified (usually 5671), as the default of 5672 is
@@ -87,7 +75,12 @@ class Connection(AbstractChannel):
         # properties set in the Tune method
         self.channel_max = channel_max
         self.frame_max = frame_max
+        # final heartbeat interval value (in float seconds) after negotiation
+        self.heartbeat = 0
+        # original heartbeat interval value proposed by client
         self.client_heartbeat = heartbeat
+        # original heartbeat interval proposed by server
+        self.server_heartbeat = None
 
         self.publisher_ack_enabled = enable_publisher_ack
 
@@ -130,9 +123,42 @@ class Connection(AbstractChannel):
 
         self._send_open(virtual_host)
 
+        # set up automatic heartbeats
+        self._close_event = Event()
+        if auto_heartbeat:
+            log.debug('Start automatic heartbeat thread')
+            t = Thread(target=self._heartbeat_thread, name='HeartbeatThread')
+            t.start()
+
     @property
     def connected(self):
-        return self.transport and self.transport.connected
+        """Check if connection is connected
+
+        :return: True if connected, else False
+        :rtype: bool
+        """
+        return bool(self.transport and self.transport.connected)
+
+    @property
+    def sock(self):
+        """Access underlying TCP socket
+
+        :return: socket
+        :rtype: socket.socket
+        """
+        if self.transport and self.transport.sock:
+            return self.transport.sock
+
+    @property
+    def server_capabilities(self):
+        """Get server capabilities
+
+        These properties are set only after successfully connecting.
+
+        :return: server capabilities
+        :rtype: dict
+        """
+        return self.server_properties.get('capabilities') or {}
 
     @synchronized('lock')
     def channel(self, channel_id=None):
@@ -145,14 +171,111 @@ class Connection(AbstractChannel):
         """
         return self.channels.get(channel_id, Channel(self, channel_id))
 
-    def _do_close(self):
+    def send_heartbeat(self):
+        """Send a heartbeat to the server
+        """
+        self.transport.write_frame(Frame(FrameType.HEARTBEAT))
+
+    def is_alive(self):
+        """Check if connection is alive
+
+        This method is the primary way to check if the connection is alive.
+
+        Side effects: This method may send a heartbeat as a last resort to check if the connection is alive.
+
+        :return: True if connection is alive, else False
+        :rtype: bool
+        """
+        if not self.sock:
+            # we don't have a valid socket, this connection is definitely not alive
+            return False
+
+        if not self.connected:
+            # the `transport` is not connected
+            return False
+
+        # recv with MSG_PEEK to check if the connection is alive
+        # note: if there is data still in the buffer, this will not tell us anything
+        if hasattr(socket, 'MSG_PEEK'):
+            prev = self.sock.gettimeout()
+            self.sock.settimeout(0.0001)
+            try:
+                self.sock.recv(1, socket.MSG_PEEK)
+            except socket.timeout:
+                pass
+            except socket.error:
+                # the exception is usually (always?) a ConnectionResetError in Python 3.3+
+                return False
+            finally:
+                self.sock.settimeout(prev)
+
+        # send a heartbeat to check if the connection is alive
+        try:
+            self.send_heartbeat()
+        except socket.error:
+            return False
+
+        return True
+
+    def drain_events(self, timeout=None):
+        """Wait for an event on all channels
+
+        This method should be called after creating consumers in order to receive delivered messages and execute
+        consumer callbacks.
+
+        :param timeout: maximum allowed time wait for an event
+        :type timeout: float or None
+        :raise amqpy.exceptions.Timeout: if the operation times out
+        """
+        chan_id, method = self._wait_multiple(self.channels, None, timeout=timeout)
+        channel = self.channels[chan_id]
+        return self.handle_method(method, channel=channel)
+
+    def close(self, reply_code=0, reply_text='', method_type=method_t(0, 0)):
+        """Close connection to the server
+
+        This method performs a connection close handshake with the server, then closes the underlying connection.
+
+        If this connection close is due to a client error, the client may provide a `reply_code`, `reply_text`,
+        and `method_type` to indicate to the server the reason for closing the connection.
+
+        :param int reply_code: the reply code
+        :param str reply_text: localized reply text
+        :param method_type: if close is triggered by a failing method, this is the method that caused it
+        :type method_type: amqpy.spec.method_t
+        """
+        # signal to the heartbeat thread to stop sending heartbeats
+        self._close_event.set()
+
+        if not self.is_alive():
+            # already closed
+            log.debug('Already closed')
+            return
+
+        args = AMQPWriter()
+        args.write_short(reply_code)
+        args.write_shortstr(reply_text)
+        args.write_short(method_type.class_id)
+        args.write_short(method_type.method_id)
+        self._send_method(Method(spec.Connection.Close, args))
+        return self.wait(allowed_methods=[spec.Connection.Close, spec.Connection.CloseOk])
+
+    def _heartbeat_thread(self):
+        # `is_alive()` sends heartbeats if the connection is alive
+        while self.is_alive():
+            # `close` is set to true if the `close_event` is signalled
+            close = self._close_event.wait(self.heartbeat / 1.5)
+            if close:
+                break
+
+    def _close(self):
         try:
             self.transport.close()
 
             channels = [x for x in self.channels.values() if x is not self]
             for ch in channels:
                 # noinspection PyProtectedMember
-                ch._do_close()
+                ch._close()
         except socket.error:
             pass  # connection already closed on the other end
         finally:
@@ -181,60 +304,8 @@ class Connection(AbstractChannel):
         except ValueError:
             raise AMQPConnectionError('Channel {} already open'.format(channel_id))
 
-    def is_alive(self):
-        if hasattr(socket, 'MSG_PEEK'):
-            sock = self.sock
-            prev = sock.gettimeout()
-            sock.settimeout(0.0001)
-            try:
-                sock.recv(1, socket.MSG_PEEK)
-            except socket.timeout:
-                pass
-            except socket.error:
-                return False
-            finally:
-                sock.settimeout(prev)
-        return True
-
-    def drain_events(self, timeout=None):
-        """Wait for an event on a channel
-
-        :param timeout: maximum allowed time wait for an event
-        :type timeout: float or None
-        :raise amqpy.exceptions.Timeout: if the operation times out
-        """
-        chan_id, method = self._wait_multiple(self.channels, None, timeout=timeout)
-        channel = self.channels[chan_id]
-        return self.handle_method(method, channel=channel)
-
-    def close(self, reply_code=0, reply_text='', method_type=method_t(0, 0)):
-        """Request a connection close
-
-        This method indicates that the sender wants to close the connection. This may be due to internal conditions
-        (e.g. a forced shut-down) or due to an error handling a specific method, i.e. an exception. When a close is due
-        to an exception, the sender provides the class and method id of the method which caused the exception.
-
-        :param reply_code: the reply code
-        :param reply_text: localized reply text
-        :param method_type: if close is triggered by a failing method, this is the method that caused it
-        :type reply_code: int
-        :type reply_text: str
-        :type method_type: amqpy.spec.method_t
-        """
-        if self.transport is None:
-            # already closed
-            return
-
-        args = AMQPWriter()
-        args.write_short(reply_code)
-        args.write_shortstr(reply_text)
-        args.write_short(method_type.class_id)
-        args.write_short(method_type.method_id)
-        self._send_method(Method(spec.Connection.Close, args))
-        return self.wait(allowed_methods=[spec.Connection.Close, spec.Connection.CloseOk])
-
     def _cb_close(self, method):
-        """Respond to a connection close
+        """Handle received connection close
 
         This method indicates that the sender (server) wants to close the connection. This may be due to internal
         conditions (e.g. a forced shut-down) or due to an error handling a specific method, i.e. an exception. When a
@@ -274,7 +345,7 @@ class Connection(AbstractChannel):
         Close-Ok handshake method SHOULD log the error.
         """
         self._send_method(Method(spec.Connection.CloseOk))
-        self._do_close()
+        self._close()
 
     def _cb_close_ok(self, method):
         """Confirm a connection close
@@ -283,7 +354,7 @@ class Connection(AbstractChannel):
         close the underlying connection.
         """
         assert method
-        self._do_close()
+        self._close()
 
     def _send_open(self, virtual_host, capabilities=''):
         """Open connection to virtual host
@@ -434,7 +505,7 @@ class Connection(AbstractChannel):
         self._send_method(Method(spec.Connection.StartOk, args))
 
     def _cb_tune(self, method):
-        """Propose connection tuning parameters
+        """Handle received "tune" method
 
         This method is the handler for receiving a "tune" method. `channel_max` and `frame_max` are set to the lower
         of the values proposed by each party.
@@ -465,6 +536,7 @@ class Connection(AbstractChannel):
         # largest frame size the server proposes for the connection
         self.frame_max = min(args.read_long(), self.frame_max)
         self.method_writer.frame_max = self.frame_max
+        # heartbeat interval proposed by server
         self.server_heartbeat = args.read_short() or 0
 
         # negotiate the heartbeat interval to the smaller of the specified values
@@ -479,9 +551,6 @@ class Connection(AbstractChannel):
 
         self._send_tune_ok(self.channel_max, self.frame_max, self.heartbeat)
 
-    def send_heartbeat(self):
-        self.transport.write_frame(Frame(FrameType.HEARTBEAT))
-
     def _send_tune_ok(self, channel_max, frame_max, heartbeat):
         """Negotiate connection tuning parameters
 
@@ -516,14 +585,6 @@ class Connection(AbstractChannel):
         self._send_method(Method(spec.Connection.TuneOk, args))
         self._wait_tune_ok = False
 
-    @property
-    def sock(self):
-        return self.transport.sock
-
-    @property
-    def server_capabilities(self):
-        return self.server_properties.get('capabilities') or {}
-
     METHOD_MAP = {
         spec.Connection.Start: _cb_start,
         spec.Connection.Secure: _cb_secure,
